module Encryptor(data_set,clock,reset,key,encrypt_data);



input [7:0] data_set [47:0];  //INPUT DATA SET - 3 SETS OF 16 CHARACTERS
input clock;
input reset;
input [7:0] key [15:0];      //KEY
output [7:0] encrypt_data [47:0]; //OUTPUT DATA

reg key_gen_done;
reg [7:0] keys_out    [9:0][3:0][3:0];
reg [7:0] addr [3:0];
reg [7:0] data [3:0];
reg [7:0] key_matrix    [3:0][3:0];
reg [7:0] state_matrix0 [3:0][3:0];
reg [7:0] state_matrix1 [3:0][3:0];
reg [7:0] d0_out        [3:0][3:0]; 
reg [7:0] d1_out        [3:0][3:0]; 
reg [7:0] d2_out        [3:0][3:0]; 
reg [7:0] d3_out        [3:0][3:0]; 
reg [7:0] d4_out        [3:0][3:0];  
reg [7:0] d5_out        [3:0][3:0]; 
reg [7:0] d6_out        [3:0][3:0]; 
reg [7:0] d7_out        [3:0][3:0];
reg [7:0] d8_out        [3:0][3:0]; 
reg [7:0] d9_out        [3:0][3:0];  

reg d_done[9:0];
reg enable_4;

reg new_word;

reg [7:0] encrypt_out [47:0];

reg enable;
reg [5:0] state;
reg [4:0] i,j,k,n,n1;
parameter key_expansion = 4'd0,get_data=4'd1, init_add_round_key = 4'd2, wait_state = 4'd3,transfer = 4'd4,next_word=4'd5;

//KEY GENERATOR MODULE	 
key_generator K(
						.clk(clock),
						.reset(reset),
						.initial_key(key_matrix),
						.keys(keys_out),
						.key_gen_done(key_gen_done)
					);
					

//10 ROUNDS OF ENCRYPTION
encryption_single d0(   
                    .data_set(state_matrix1),
                    .clock(clock),
						  .reset(reset | ~key_gen_done | new_word ),
						  .expanded_key(keys_out[0]),
						  .round_num(4'd1),
						  .round_data(d0_out),
						  .round_over(d_done[0])

                  ); 
 
 encryption_single d1(   
                    .data_set(d0_out),
                    .clock(clock),
						  .reset(d_done[0]),
						  .expanded_key(keys_out[1]),
						  .round_num(4'd2),
						  .round_data(d1_out),
						  .round_over(d_done[1])

                  );
					
 encryption_single d2(   
                    .data_set(d1_out),
                    .clock(clock),
						  .reset(d_done[1]),
						  .expanded_key(keys_out[2]),
						  .round_num(4'd3),
						  .round_data(d2_out),
						  .round_over(d_done[2])

                  ); 
							
 encryption_single d3(   
                    .data_set(d2_out),
                    .clock(clock),
						  .reset(d_done[2]),
						  .expanded_key(keys_out[3]),
						  .round_num(4'd4),
						  .round_data(d3_out),
						  .round_over(d_done[3])

                  ); 
						
 encryption_single d4(   
                    .data_set(d3_out),
                    .clock(clock),
						  .reset(d_done[3]),
						  .expanded_key(keys_out[4]),
						  .round_num(4'd5),
						  .round_data(d4_out),
						  .round_over(d_done[4])

                  ); 
						

						

						
 encryption_single d5(   
                    .data_set(d4_out),
                    .clock(clock),
						  .reset(d_done[4]),
						  .expanded_key(keys_out[5]),
						  .round_num(4'd6),
						  .round_data(d5_out),
						  .round_over(d_done[5])

                  ); 
						
 encryption_single d6(   
                    .data_set(d5_out),
                    .clock(clock),
						  .reset(d_done[5]),
						  .expanded_key(keys_out[6]),
						  .round_num(4'd7),
						  .round_data(d6_out),
						  .round_over(d_done[6])

                  ); 
						
						
 encryption_single d7(   
                    .data_set(d6_out),
                    .clock(clock),
						  .reset(d_done[6]),
						  .expanded_key(keys_out[7]),
						  .round_num(4'd8),
						  .round_data(d7_out),
						  .round_over(d_done[7])

                  ); 
						
 encryption_single d8(   
                    .data_set(d7_out),
                    .clock(clock),
						  .reset(d_done[7]),
						  .expanded_key(keys_out[8]),
						  .round_num(4'd9),
						  .round_data(d8_out),
						  .round_over(d_done[8])

                  ); 
						
 encryption_single d9(   
                    .data_set(d8_out),
                    .clock(clock),
						  .reset(d_done[8]),
						  .expanded_key(keys_out[9]),
						  .round_num(4'd10),
						  .round_data(d9_out),
						  .round_over(d_done[9])

                  ); 
									
					
always@(posedge clock) begin
if(reset) begin
	enable<=0;
	state<= key_expansion;
		for(i=0;i<4;i=i+1) begin
		   for(j=0;j<4;j=j+1)begin
			key_matrix[j][i] <= key[i*4 + j];
		  end
	   end
	n1 <= 0;
	n <= 0;
	new_word <= 0;
end

//WAIT FOR ALL KEYS TO BE GENERATED
else begin
   case(state)
	 key_expansion: begin
	   if(key_gen_done == 1) begin
	     state <= get_data;
		end
	  end
//GET THE INPUT DATA (16 CHARACTERS AT A TIME)
	  get_data: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix0[j][i] <= data_set[n*16 + i*4 + j];
				end
			end
			state <= init_add_round_key;
		end

//PEROFRM INITIAL OPERATION : XOR WITH INITIAL KEY		
	  init_add_round_key: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix1[j][i] <= state_matrix0[j][i] ^ key_matrix[j][i];
				end
			end
		state <= wait_state;
		new_word <= 0;
		end


//WAIT IN THIS STATE TILL FIRST ROUND IS COMPLETED SO THAT YOU CAN GO ABCK AND START ENCRYPTION FOR NEXT SET OF DATA.
//THIS HELPS US IN PIPELINING OUR OPERATIONS		
		wait_state: begin
			if(d_done[0] == 1) begin
				if(n < 3) begin
				new_word <= 1;
				state <= get_data;
				n <= n+1;
				end
			end
			else begin
				state <= wait_state;
			end

//TRANSFER THE RESULTS OF LAST ROUND TO THE OUTPUT		
			if(d_done[9] == 1) begin
		if(n1>0) begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					encrypt_out[(n1-1)*16 + i*4 + j] <= d9_out[j][i];
				end
			end
		end
		n1 <= n1+1;
	 end
		end
		
	endcase
end
end



assign encrypt_data = encrypt_out;
	
endmodule